/*                                                                              */
/* Copyright (C) 2001 International Business Machines                           */
/* All rights reserved.                                                         */
/*                                                                              */
/* This file is part of the GPFS user library.                                  */
/*                                                                              */
/* Redistribution and use in source and binary forms, with or without           */
/* modification, are permitted provided that the following conditions           */
/* are met:                                                                     */
/*                                                                              */
/*  1. Redistributions of source code must retain the above copyright notice,   */
/*     this list of conditions and the following disclaimer.                    */
/*  2. Redistributions in binary form must reproduce the above copyright        */
/*     notice, this list of conditions and the following disclaimer in the      */
/*     documentation and/or other materials provided with the distribution.     */
/*  3. The name of the author may not be used to endorse or promote products    */
/*     derived from this software without specific prior written                */
/*     permission.                                                              */
/*                                                                              */
/* THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR         */
/* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES    */
/* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.      */
/* IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, */
/* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;  */
/* OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,     */
/* WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR      */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF       */
/* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                   */
/*                                                                              */
/* @(#)39       1.6  src/avs/fs/mmfs/ts/util/gpfs_lweTypes.h, mmfs, avs_rfks1, rfks1s006a 7/14/14 08:35:13 */
/*
 *  Library calls for GPFS interfaces
 */
#ifndef _h_lwe_types
#define _h_lwe_types

/* LWE Event Types */
typedef enum 
{

  LWE_EVENT_FILEOPEN_READ   = 0, /* Open for Read Only -  EVENT 'OPEN_READ' */
  LWE_EVENT_FILEOPEN_WRITE  = 1, /* Open with Writing privileges - EVENT 'OPEN_WRITE' */
  LWE_EVENT_FILECLOSE       = 2, /* "File Close Event" 'CLOSE' */
  LWE_EVENT_FILEREAD        = 3, /* "File Read Event" 'READ' */
  LWE_EVENT_FILEWRITE       = 4, /* "File Write Event" 'WRITE' */
  LWE_EVENT_FILEDESTROY     = 5, /* File is being destroyed 'DESTROY' */
  LWE_EVENT_FILEEVICT       = 6, /* OpenFile object is being evicted from memory 'FILE_EVICT' */
  LWE_EVENT_BUFFERFLUSH     = 7, /* Data buffer is being written to disk 'BUFFER_FLUSH' */
  LWE_EVENT_POOLTHRESHOLD   = 8, /* Storage pool exceeded defined utilization 'POOL_THRESHOLD' */
  /* support 32 events in our 32 bit lweEventMask in the StripeGroup */
  /* must coordinate with eventNameToType & co in lwe.C */
  LWE_EVENT_UNKNOWN         = 32,/* "Uknown event" */
} lwe_event_type;

/* LWE event resposne type */
typedef enum
{
  LWE_RESP_PENDING  = 0,  /* "Response Unknown" */
  LWE_RESP_CONTINUE = 1,  /* "Response Continue" */
  LWE_RESP_ABORT    = 2,  /* "Response ABORT" */
  LWE_RESP_DONTCARE = 3   /* "Response DONTCARE" */
} lwe_resp_t;

#define lwe_event_type unsigned int
#define lwe_event_token unsigned long long
#define lwe_token_t unsigned long long
#define lwe_sessid_t unsigned int

/* lwe event structure, for external interfance */
typedef struct lwe_event_s {
  int              eventLen;        /* offset 0 */
  lwe_event_type   eventType;       /* offset 4 */
  lwe_event_token  eventToken;      /* offset 8 <--- Must on DWORD */
  int              isSync;          /* offset 16 */
  int              parmLen;         /* offset 20 */
  char*            parmP;           /* offset 24 <-- Must on DWORD */
} lwe_event_t;

#define MAX_LWE_SESSIONS 1024
#define MAX_LWE_EVENTS   1024
#define MAX_LWESESSION_INFO_LEN 100

#define LWE_NO_SESSION 0
#define LWE_EV_NOWAIT 1

#endif /* _h_lwe_types */
