/* IBM_PROLOG_BEGIN_TAG                                                   */
/* This is an automatically generated prolog.                             */
/*                                                                        */
/*                                                                        */
/*                                                                        */
/* Licensed Materials - Property of IBM                                   */
/*                                                                        */
/* (C) COPYRIGHT International Business Machines Corp. 2003,2014          */
/* All Rights Reserved                                                    */
/*                                                                        */
/* US Government Users Restricted Rights - Use, duplication or            */
/* disclosure restricted by GSA ADP Schedule Contract with IBM Corp.      */
/*                                                                        */
/* IBM_PROLOG_END_TAG                                                     */
static char sccsid[] = "@(#)34	1.18  src/avs/fs/mmfs/samples/util/tsfindinode.c, mmfs, avs_rfks1, rfks1s006a 2/20/14 10:28:31";

/*====================================================================
 *
 * tsfindinode command: Recursively traverse the directory structure
 *                    to list all files/directories in the file system
 *                    that have any of the inode numbers listed.
 *
 * Syntax: tsfindinode {-i {inum | inumfile}} [-0] [-t nThreads] [-a fsName] [--] <pathname>
 *
 *   The "pathname" parameter should be the directory where the root of
 *   the file system is found.
 *
 *==================================================================*/

#ifdef GPFS_LINUX
/* Use 64 bit version of stat, etc. */
#define _LARGEFILE_SOURCE
#define _FILE_OFFSET_BITS 64

typedef long long offset_t;
typedef long long InodeCount;
#else
/* Only 64 bit inode numbers and > 2B inodes supported as of 3.5 */
typedef long long InodeCount;
#endif

#ifdef GPFS_AIX
/* Use 64 bit version of stat, etc. */
#ifndef _LARGE_FILES
#define _LARGE_FILES
#endif
#endif

#ifdef GW_POSIX
#include <gpfsWinPosix.h>
#endif

#define GPFS_64BIT_INODES 1

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <libgen.h>
#include <pthread.h>
#include <sys/wait.h>

#include <gpfs.h>

int optcheck = 1;
char *RootFsDirP;

/* Below is used to scan all snapshots including global and
   fileset snapshots. */
#define MAX_NAME_LIST_INCREMENT  50
typedef enum name_item_type
{
  SNAPSHOT_ITEM,
  FILESET_ITEM,
  WORKINGFILE_ITEM,
}NameItemType;

typedef enum list_action_type
{
  SNAPSHOT_LIST,
  FILESET_LIST,
  FILESET_STATUS,
}ListActionType;

/* Use to save all snapshots, or all filesets, or working files. */
typedef struct name_list_item
{
  long long id;
  ino_t fsetRootIno;
  char name[1024];
  FILE *fileP;
} NameListItem;

NameListItem *snapListP = NULL;
NameListItem *filesetListP = NULL;
NameListItem *workingFileListP = NULL;
int maxSnapItems = 0, maxFilesetItems = 0, maxWorkingFileItems = 0;
int numSnapItems = 0, numFilesetItems = 0, numWorkingFileItems = 0;
char* fsName = NULL;
char* inputFileName = NULL;
static int isScanAllSnapshots = 0;
static int snapIdIndex = -1;
static ino_t rootInoOfRootFset = 3;
/* End of newly added for all snapshots scan. */

/* Define the number of threads used to read the directories */
#define DEFAULT_THREADS 16
int nThreads = DEFAULT_THREADS;
const char *print0 = "\n";

/* Define wait queue for threads */
static pthread_mutex_t QueueMutex;
static pthread_cond_t QueueCond;
static int NWorkersWaiting = 0;
static int NWorkersRunning = 0;

typedef struct queue_element 
{
  struct queue_element *nextP;
  ino_t ino;
  char path[4];
} QueueElement; 

static QueueElement *WorkQueueP = NULL;

/* Define mutex to serialize the output */
static pthread_mutex_t OutputMutex;

#define MAX_INODE_INCREMENT 1000
static InodeCount numInodes = 0;
static InodeCount maxInodes = 0;
static ino_t *Inode = NULL;
static unsigned char *InodeSeen = NULL;

/* error code set while in threads to control final error code */
int exitrc = 0;

/* print usage message and exit */
void usage(char *argv0)
{
  fprintf(stderr,
     "Usage: %s {-i {inum | inumfile}} [-0] [-t nThreads (def=%d)] [-a fsName] [--] pathname\n"
     "       (-i can be used multiple times to specify another inode number or file of inode numbers)\n"
     "       (-a with fs name and -i with inumfile are needed to scan all snapshots, used by mmfileid)\n",
          basename(argv0), DEFAULT_THREADS);
  exit(1);
}

/* Below new functions are used to scan all snapshots, 
   currently mmfileid use it only.*/
int setFilesetStatus(char* name, char* status)
{
  NameListItem *headP = filesetListP;
  int rc = ENOENT;
  int index;
  long long fsetStatus;

  if (strcmp(status, "Linked") == 0)
    fsetStatus = 0x0000000100000000LL;
  else
    fsetStatus = 0x0;

  for (index = 0; index < numFilesetItems; index ++)
  {
    if (strcmp(name, headP[index].name) == 0)
    {
      rc = 0;
      headP[index].id = headP[index].id | fsetStatus;
      break;
    }
  }

  if (rc != 0)
    fprintf(stderr, "Can not find the fileset '%s' in the list\n", name);
  
  return rc;
}

NameListItem* getSnapOrFilesetItem(int id, int isSnapshot)
{
  NameListItem *headP;
  int i, low, high;
  long long searchKey = id;
  long long itemKey;

  if (isSnapshot)
  {
    headP = snapListP;
    high = numSnapItems - 1;
  }
  else
  {
    headP = filesetListP;
    high = numFilesetItems - 1;
  }

  /* Fast check. */
  if (headP == NULL)
    return NULL;

  /* Binary search on the list. */
  low = 0;
  while (low <= high)
  {
    i = (low + high) / 2;
    itemKey = headP[i].id & 0x00000000FFFFFFFF;
    if (searchKey == itemKey)
      return (headP + i);
    if (low == high)
      return NULL;
    if (searchKey < itemKey)
      high = i - 1;
    else
      low = i + 1;
  }

  return NULL;
}

/* Add new item into snapshot, fileset or working file list.
   The 'itemType' parameter indicates which list the new item
   will be insert into. */
int addNameItem(NameItemType itemType, ino_t ino,
                long long id, int isGlobal,
                char* snapName, char* fsetName)
{
  NameListItem *headP, *tmpP;
  gpfs_fssnap_handle_t *fsetP = NULL;
  char* snapDirPath = NULL;
  char workingFileName[1024];
  int numItems, maxItems;
  int i, j, low, high;
  int snapId, fsetId, rc = 0;

  if (itemType == SNAPSHOT_ITEM)
  {
    numItems = numSnapItems;
    maxItems = maxSnapItems;
    headP = snapListP;
  }
  else if (itemType == FILESET_ITEM)
  {
    numItems = numFilesetItems;
    maxItems = maxFilesetItems;
    headP = filesetListP;
  }
  else /* WORKINGFILE_IEM. */
  {
    numItems = numWorkingFileItems;
    maxItems = maxWorkingFileItems;
    headP = workingFileListP;
  }
  if (headP == NULL || numItems >= maxItems)
  {
    tmpP = (NameListItem*)malloc(sizeof(NameListItem) *
              (numItems + MAX_NAME_LIST_INCREMENT));
    if (tmpP == NULL)
    {
      fprintf(stderr, "Failed to alloc NameList\n");
      return ENOMEM;
    }

    /* Init the new list. */
    if (headP != NULL)
    {
      memcpy(tmpP, headP, numItems * sizeof(NameListItem));
      free(headP);
    }
    maxItems = maxItems + MAX_NAME_LIST_INCREMENT;
    headP = tmpP;
    if (itemType == SNAPSHOT_ITEM)
    {
      snapListP = headP;
      maxSnapItems = maxItems;
    }
    else if (itemType == FILESET_ITEM)
    {
      filesetListP = headP;
      maxFilesetItems = maxItems;
    }
    else
    {
      workingFileListP = headP;
      maxWorkingFileItems = maxItems;
    }
  }

  if (numItems == 0)
    j = 0;
  else
  {
    /* Binary search the list. */
    i = 0;
    low = 0;
    high = numItems - 1;
    while (low <= high)
    {
      i = (low + high) / 2;
      /* Ignore duplicated items. */
      if (id == headP[i].id)
      {
        if (itemType == WORKINGFILE_ITEM)
        {
          if (fprintf(headP[i].fileP, "%lld\n", ino) < 0)
          {
            rc = errno;
            fprintf(stderr, "Error on writing inode %lld; %s\n",
                    ino, strerror(rc));
          }
        }
        goto exit;
      }
      if (low == high)
      {
        if (id > headP[i].id)
          i++;
        break;
      }
      if (id < headP[i].id)
        high = i - 1;
      else
        low = i + 1;
    }

    /* The output of mmlssnapshot and mmlsfileset is already sorted
       by id, below move seems only needed for working file list.
       But for safe, also do move for snap and fileset list. */
    for (j = numItems; j > i; j--)
      headP[j] = headP[j-1];
  }

  /* Insert the new entry. */
  numItems ++;
  headP[j].id = id;
  headP[j].fsetRootIno = ino;
  if (itemType != WORKINGFILE_ITEM)
    strcpy(headP[j].name, snapName);
  else
  {
    /* Get the snapdir. */
    if (isGlobal)
      fsetP = gpfs_get_fssnaphandle_by_name(fsName, snapName);
    else
      fsetP = gpfs_get_fset_snaphandle_by_name(fsName, fsetName, snapName);
    if (fsetP == NULL)
    {
      rc = errno;
      fprintf(stderr, "failed to %s,"
              " fsName: %s snapName: %s fsetName: %s isGlobal: %d, %s\n",
              isGlobal ? "gpfs_get_fssnaphandle_by_name" :
                         "gpfs_get_fset_snaphandle_by_name",
              fsName, snapName == NULL ? "NULL" : snapName,
              fsetName == NULL ? "NULL" : fsetName, isGlobal,
              strerror(rc));
      goto exit;
    }
    else
    {
      snapDirPath = (char*)gpfs_get_pathname_from_fssnaphandle(fsetP);
      if (snapDirPath == NULL)
      {
        rc = errno;
        fprintf(stderr, "failed to gpfs_get_pathname_from_fssnaphandle, %s\n",
                strerror(rc));
        goto exit;
      }
      else
        strcpy(headP[j].name, snapDirPath);
    }

    /* Open the file, write the inode. */
    snapId = id >> 32;
    fsetId = id & 0xFFFF;
    snprintf(workingFileName, sizeof(workingFileName), "%s_%d_%d\0", 
             inputFileName, snapId, fsetId);
    headP[j].fileP = fopen(workingFileName, "w");
    if (headP[j].fileP == NULL)
    {
      rc = errno;
      fprintf(stderr, "Error on open file %s; %s\n",
              workingFileName, strerror(rc));
    }
    if (fprintf(headP[j].fileP, "%lld\n", ino) < 0)
    {
      rc = errno;
      fprintf(stderr, "Error on writing inode %lld to file %s; %s\n",
              ino, workingFileName, strerror(rc));
      goto exit;
    }
  }

  if (itemType == SNAPSHOT_ITEM)
    numSnapItems = numItems;
  else if (itemType == FILESET_ITEM)
    numFilesetItems = numItems;
  else
    numWorkingFileItems = numItems;

exit:
  if (fsetP)
    gpfs_free_fssnaphandle(fsetP);
  return rc;
}

/* Group inodes from the input inode file into working files, each
   working file has the same snap/fset id. */
int constructWorkingFiles()
{
  NameListItem *tempItemP;
  FILE *inputFd;
  ino_t ino;
  long long key;
  int snapId, fsetId, isGlobalSnap;
  long long diskId, sectorId;
  char lineBuffer[1024];
  char *snapName = NULL, *fsetName = NULL;
  int index, rc;

  inputFd = fopen(inputFileName, "r");
  if (inputFd == NULL)
  {
    rc = errno;
    fprintf(stderr, "Error opening %s; %s\n", inputFileName, strerror(rc));
    return rc;
  }

  memset(lineBuffer, 0, sizeof(lineBuffer));
  while (fgets(lineBuffer, sizeof(lineBuffer), inputFd) != NULL)
  {
    if (sscanf(lineBuffer, "%lld %lld:%lld %d %d %d", 
               &ino, &diskId, &sectorId, &snapId, &isGlobalSnap, &fsetId) == 6)
    {
      /* Get snap and fset name. */
      tempItemP = getSnapOrFilesetItem(snapId, 1);
      snapName = tempItemP ? tempItemP->name : NULL;
      tempItemP = getSnapOrFilesetItem(fsetId, 0);
      fsetName = tempItemP ? tempItemP->name : NULL;
      if ((snapName == NULL && (snapId != 0 || !isGlobalSnap)) || 
          fsetName == NULL)
      {
        fprintf(stderr, "Invalid data: %s\n", lineBuffer);
        memset(lineBuffer, 0, sizeof(lineBuffer));
        continue;
      }

      /* rootInode of the fileset in fileset snapshot will not be 
         found by gpfs_ireaddir. */
      if (!isGlobalSnap && ino == tempItemP->fsetRootIno)
      {
        fprintf(stderr,
                "%10llu\t%10d\t(rootInode of fileset '%s' in snapshot '%s')\n",
                ino, snapId, fsetName, snapName);
        memset(lineBuffer, 0, sizeof(lineBuffer));
        continue;
      }
   
      /* rootInode of 'root' fileset in global snapshot will not be
         found by gpfs_ireaddir. */
      if (snapId != 0 && isGlobalSnap)
      {
        if (ino == rootInoOfRootFset)
        {
          fprintf(stderr,
                  "%10llu\t%10d\t(rootInode of fileset '%s' in snapshot '%s')\n",
                  ino, snapId, fsetName, snapName);
          memset(lineBuffer, 0, sizeof(lineBuffer));
          continue;
        }
      }

      /* If the inode in active file system or a fileset snapshot, the
         fileset need to be linked, else we gpfs_ireaddir can not find it. */
      if (snapId == 0 || !isGlobalSnap)
      {
        if ((tempItemP->id & 0x0000000100000000LL) == 0x0)
        {
          fprintf(stderr, "%10llu\t%10d\t(notfound, fileset '%s' is unlinked)\n",
                  ino, snapId, fsetName);
          memset(lineBuffer, 0, sizeof(lineBuffer));
          continue;
        }
      }

      /* Search or insert new item into workingFileListP. */
      key = snapId;
      key = (key << 32) | fsetId;
      rc = addNameItem(WORKINGFILE_ITEM, ino, key, isGlobalSnap, snapName, fsetName);
      if (rc != 0)
        break;
    }
    else
      fprintf(stderr, "Invalid data: %s\n", lineBuffer);

    memset(lineBuffer, 0, sizeof(lineBuffer));
  }
  fclose(inputFd);

  /* Close all opened working files. */
  for (index = 0; index < numWorkingFileItems; index ++)
  {
    tempItemP = workingFileListP + index;
    if (fclose(tempItemP->fileP) < 0)
    {
      if (rc == 0)
        rc = errno;
      fprintf(stderr, "Error on closing working file for snapId %d fsetId %d; %s\n",
              tempItemP->id >> 32, tempItemP->id & 0xFFFF, strerror(rc));
    }
  }

  return rc;
}

int listSnapOrFilesetNames(ListActionType actionType)
{
  FILE *fileP = NULL;
  int started = 0;
  char cmdString[256];
  char name[256];
  char status[256];
  char dataBuffer[1024];
  ino_t fsetRootIno = 0;
  int id;
  int len;
  int rc = 0;

  if (actionType == SNAPSHOT_LIST)
    snprintf(cmdString, sizeof(cmdString),
             "/usr/lpp/mmfs/bin/mmlssnapshot %s 2>/dev/null \n",
             fsName);
  else if (actionType == FILESET_LIST)
    snprintf(cmdString, sizeof(cmdString),
             "/usr/lpp/mmfs/bin/mmlsfileset %s -L 2>/dev/null \n",
             fsName);
  else
    snprintf(cmdString, sizeof(cmdString),
             "/usr/lpp/mmfs/bin/mmlsfileset %s 2>/dev/null \n",
             fsName);
  fileP = popen(cmdString, "r");
  if (fileP == NULL)
  {
    rc = errno;
    goto exit;
  }

  while (fgets(dataBuffer, sizeof(dataBuffer), fileP) != NULL)
  {
    /* Check to start the following start. */
    if (!started)
    {
      if (actionType == SNAPSHOT_LIST  && strstr(dataBuffer, "Directory") != NULL)
        started = 1;
      else if (actionType == FILESET_LIST && strstr(dataBuffer, "RootInode") != NULL)
        started = 1;
      else if (actionType == FILESET_STATUS && strstr(dataBuffer, "Status") != NULL)
        started = 1;
      continue;
    }

    if (actionType == SNAPSHOT_LIST ||
        actionType == FILESET_LIST)
    {
      /* Scan the dataBuffer to get snapName and snapId. */
      if (actionType == SNAPSHOT_LIST)
      {
        if (sscanf(dataBuffer, "%s %d", name, &id) != 2)
        {
          fprintf(stderr, "Invalid data format: %s\n", dataBuffer);
          continue;
        }
      }
      else
      {
        if ((rc = sscanf(dataBuffer, "%s %d %lld", name, &id, &fsetRootIno)) != 3)
        {
          if (rc < 2)
          {
            fprintf(stderr, "Invalid data format: %s\n", dataBuffer);
            continue;
          }
          else
          {
            /* The fileset is deleted, but we still need the name. */
            len = strlen(name);  
            name[len - 1] = '\0';
            memmove(name, name + 1, len - 2);
            fsetRootIno = -1;
          }
        }
        if (strcmp(name, "root") == 0)
          rootInoOfRootFset = fsetRootIno;
      }

      /* Build the snap or fileset list. */
      rc = addNameItem(actionType == SNAPSHOT_LIST ? SNAPSHOT_ITEM : FILESET_ITEM, 
                       fsetRootIno, id, 0, name, NULL);
      if (rc != 0)
        break;
    }
    else
    {
      /* To get filesetname and status. */
      if (sscanf(dataBuffer, "%s %s", name, status) != 2)
      {
        fprintf(stderr, "Invalid data format: %s\n", dataBuffer);
      }
      if (strcmp(status, "Deleted") == 0)
      {
        /* The fileset is deleted, but we still need the name. */
        len = strlen(name);  
        name[len - 1] = '\0';
        memmove(name, name + 1, len - 2);
      }
      rc = setFilesetStatus(name, status);
      if (rc != 0)
        break;
    }
  }

exit:

  if (fileP)
    pclose(fileP);

  return rc;
}

/* Get the snapshot list by using mmlssnapshot. */
int constructSnapshotList()
{
  return listSnapOrFilesetNames(SNAPSHOT_LIST);
}

/* Get the fileset list by using mmlsfileset. */
int constructFilesetList()
{
  return listSnapOrFilesetNames(FILESET_LIST);
}

int constructFilesetStatusList()
{
  return listSnapOrFilesetNames(FILESET_STATUS);
}

/* Unlink working all files, free allocated memory. */
void scanAllSnapshotsPostprocess()
{
  char workingFileName[1024];
  int index, snapId, fsetId, rc;
  NameListItem* tempItemP;

  free(filesetListP);
  filesetListP = NULL;
  free(snapListP);
  snapListP = NULL;

  /* Unlink all working files. */
  for (index = 0; index < numWorkingFileItems; index ++)
  {
    memset(workingFileName, 0, sizeof(workingFileName));
    tempItemP = workingFileListP + index;
    snapId = tempItemP->id >> 32;
    fsetId = tempItemP->id & 0xFFFF;
    snprintf(workingFileName, sizeof(workingFileName), "%s_%d_%d", inputFileName, snapId, fsetId);
    rc = unlink(workingFileName);
    if (rc != 0)
      fprintf(stderr, "Error on unlink working file %s; %s\n",
              workingFileName, strerror(rc));
  }

  free(workingFileListP);
  workingFileListP = NULL;
}

/* Scan necessary snapshots (global and fileset snapshot).
   1. Get all snapshot list.
   2. Get all fileset list.
   3. Read inodes from the input ino file, group the indoes
      by snap and fset id into temp working files. Get the
      snapdir for each group.
   4. For each working file, fork a new tsfindinode to do
      the real work. */
int scanAllSnapshots()
{
  int rc, index, childStatus = 0;
  pid_t pid = 0, wPid = 0;
  char cmdString[2048];
  int snapId, fsetId;
  NameListItem* tempItemP;

  /* Construct snapshot list. */
  rc = constructSnapshotList();
  if (rc != 0)
    goto exit;

  /* Construct fileset list. */
  rc = constructFilesetList();
  if (rc != 0)
    goto exit;

  /* Set fileset linked/unlinked status. */
  rc = constructFilesetStatusList();
  if (rc != 0)
    goto exit;

  /* Construct temproary working files. */
  rc = constructWorkingFiles();
  if (rc != 0)
    goto exit;

  for (index = 0; index < numWorkingFileItems; index ++)
  {
    tempItemP = workingFileListP + index;
    memset(cmdString, 0, sizeof(cmdString));
    snapId = tempItemP->id >> 32;
    fsetId = tempItemP->id & 0xFFFF;
    snprintf(cmdString, sizeof(cmdString), 
             "/usr/lpp/mmfs/bin/tsfindinode -i %s_%d_%d -t %d -s %d -- %s",
             inputFileName, snapId, fsetId, nThreads, snapId, tempItemP->name);

    /* Fork new tsfindinode to do real work in each snapshot. */
    pid = fork();
    if (pid < 0)
    {
      rc = errno;
      fprintf(stderr, "Error on forking working process; %s\n", strerror(rc));
      break;
    }
    else if (pid == 0)
    {
      /* Executed by child. */
      rc = execl("/bin/sh", "sh", "-c", cmdString, (char *)0);
      exit(errno);
    }
    else
    {
      /* Executed by parent. */
      wPid = waitpid(pid, &childStatus, 0);
      if (wPid < 0)
      {
        rc = errno;
        fprintf(stderr, "Error on waiting child; %s\n", strerror(rc));
      }
      else
        rc = WEXITSTATUS(childStatus);

      if (rc != 0)
        break;
    }
  }

exit:

  scanAllSnapshotsPostprocess();
  return rc;
}

/* ------------- end of new functions to scan all snapshots -------- */

void addInum(ino_t inode)
{
  InodeCount i,j;
  InodeCount low, high;
  ino_t *newInode = NULL;

  if (numInodes >= maxInodes)
  {
    newInode = (ino_t *)malloc((maxInodes+MAX_INODE_INCREMENT) * sizeof(ino_t));
    if (newInode == NULL)
    {
      fprintf(stderr, "Too many inodes (%lld)\n", maxInodes);
      exit(1);
    }
    if (Inode != NULL)
    {
      /* Copy old list of inodes and delete the old array */
      memcpy(newInode, Inode, maxInodes * sizeof(ino_t));
      free(Inode);
    }
    Inode = newInode;
    maxInodes += MAX_INODE_INCREMENT;
  }

  /* Binary search the list */
  i = 0;
  low = 0;
  high = numInodes - 1;
  while (low <= high)
  {
    i = (low + high)/2;
    if (inode == Inode[i])
      return;  /* ignore duplicate entry */
    if (low == high)
    {
      if (inode > Inode[i])
        i++;
      break;
    }
    if (inode < Inode[i])
      high = i - 1;
    else
      low = i + 1;
  }

  /* move all entries down one slot */
  for (j=numInodes; j > i; j--)
    Inode[j] = Inode[j-1];

  /* insert entry in sorted array */
  Inode[i] = inode;
  numInodes++;
}

void getinodes(char * listorname)
{
  FILE *fd;
  long long inum;
  char linebuf[256];
  int rc = 0;

  memset(linebuf, 0, sizeof(linebuf));

  if ((listorname[0]>='0') & (listorname[0]<='9'))
  {
    /* If arg starts with a digit, assume it is a number for sscanf to parse */
    if (sscanf(listorname, "%lld", &inum) == 1)
      addInum(inum);
  }
  else
  {
    /* Otherwise, assume the arg is a filename.
       Read inode numbers from the file until scanf fails to find a number */
    if (listorname[0] == '-')
    {
      while (scanf("%lld", &inum) == 1)
        addInum(inum);
    }
    else
    {
      fd = fopen(listorname,"r");
      if (fd != NULL)
      {
        /* while (fscanf(fd, "%lld", &inum) == 1) */
        while (fgets(linebuf, sizeof(linebuf), fd) != NULL)
        {
          if ((linebuf[0]>='0') & (linebuf[0]<='9'))
          {
            if (sscanf(linebuf, "%lld", &inum) == 1) 
              addInum(inum);
          }
          else
            fprintf(stdout, "%s\n", linebuf); /* disregard mesg or comment line */
          memset(linebuf, 0, sizeof(linebuf));
        }
        fclose(fd);
      }
      else
      {
        rc = errno;
        fprintf(stderr, "Error opening %s: %s\n", listorname, strerror(rc));
        exit(1);
      }
    }
  }
}

/* If there is no list return 1,
   or if inode is in the list return its index,
   else return -1. */
int testinode(ino_t inode)
{
  InodeCount low, mid, high;

  /* Binary search the list */
  low = 0;
  high = numInodes - 1;
  while (low <= high)
  {
    mid = (low + high)/2;
    if (inode == Inode[mid])
      return mid;
    if (low == high)
      return -1;
    if (inode < Inode[mid])
      high = mid - 1;
    else
      low = mid + 1;
  }
  return -1;
}

/* Add an element to the work queue
   and signal any thread waiting for work. */
int EnqueueWork(const char *pathP, const char *dirP, ino_t inode)
{
  int rc;
  int pathLen = strlen(pathP);
  QueueElement *qP = NULL;
  
  /* Allocate a work queue element and space for the path.  */
  qP = (QueueElement *) malloc((sizeof(*qP) - sizeof(qP->path)) 
                               + pathLen + strlen(dirP) + 2);
  if (qP == NULL)
  {
    fprintf(stderr,"Out of memory.\n");
    return(ENOMEM);
  }
  
  /* Initialize element */
  qP->ino = inode;
  if (pathLen > 0)
  {
    strcpy(qP->path, pathP);
    if (strlen(dirP) > 0)
    {
      strcat(qP->path, "/");
      strcat(qP->path, dirP);
    }
  }
  else
    strcpy(qP->path, dirP);
    
  /* Acquire mutex protecting the work queue */
  rc = pthread_mutex_lock(&QueueMutex);
  if (rc != 0)
  {
    fprintf(stderr, "Error in pthread_mutex_lock: %s\n", strerror(rc));
    return rc;
  }

  /* Add work element to the list */
  qP->nextP = WorkQueueP;
  WorkQueueP = qP;

  /* Signal a waiting worker thread */
  if (NWorkersWaiting > 0)
  {
    NWorkersWaiting -= 1;
    rc = pthread_cond_signal(&QueueCond);
    if (rc != 0)
    {
      fprintf(stderr, "Error in pthread_cond_signal: %s\n", strerror(rc));
      return rc;
    }
  }

  /* Release mutex before returning */
  rc = pthread_mutex_unlock(&QueueMutex);
  if (rc != 0)
  {
    fprintf(stderr, "Error in pthread_mutex_unlock: %s\n", strerror(rc));
    return rc;
  }
  
  /* Success! */
  return 0;
}


/* Dequeue the next work element from the queue
   or if the queue is empty, wait for more work to become available */
QueueElement *DequeueWork(void)
{
  int rc;
  QueueElement *qP;
  
  /* Acquire mutex protecting the work queue */
  rc = pthread_mutex_lock(&QueueMutex);
  if (rc != 0)
  {
    fprintf(stderr, "Error in pthread_mutex_lock: %s\n", strerror(rc));
    if (exitrc == 0)
      exitrc = rc;
    return NULL;
  }
  
  /* One less thread working */
  NWorkersRunning -= 1;

  /* Loop waiting for work to become available */
  while ((WorkQueueP == NULL) && (NWorkersRunning > 0))
  {
    NWorkersWaiting += 1;       /* One more thread waiting */
    rc = pthread_cond_wait(&QueueCond, &QueueMutex);
    if (rc != 0)
    {
      fprintf(stderr, "Error in pthread_cond_wait: %s\n", strerror(rc));
      if (exitrc == 0)
        exitrc = rc;
      return NULL;
    }
  }

  /* Check if we are done */
  if (WorkQueueP == NULL)
  {
    qP = NULL;                  /* Nothing more to do. */
    
    /* signal any waiting threads */
    if (NWorkersWaiting > 0)
    {
      NWorkersWaiting = 0;
      rc = pthread_cond_broadcast(&QueueCond);
      if (rc != 0)
      {
        fprintf(stderr, "Error in pthread_cond_broadcast: %s\n", strerror(rc));
        if (exitrc == 0)
          exitrc = rc;
      }
    }
  }
  else
  {
    /* Remove the first element from the queue */
    qP = WorkQueueP;
    WorkQueueP = qP->nextP;
    qP->nextP = NULL;

    /* One more thread working */
    NWorkersRunning += 1;
  }

  /* Release mutex before returning */
  rc = pthread_mutex_unlock(&QueueMutex);
  if (rc != 0)
  {
    fprintf(stderr, "Error in pthread_mutex_unlock: %s\n", strerror(rc));
    if (exitrc == 0)
      exitrc = rc;
    return NULL;
  }
  
  return qP;
}

/* Main body for each thread.
   Loop until there is no more work to do. */
void* ThreadBody(void *parmP)
{
  int rc;
  QueueElement *qP;
  int read_dir(const char *, ino_t);

  /* Count this thread as a worker thread */
  rc = pthread_mutex_lock(&QueueMutex);
  if (rc != 0)
  {
    fprintf(stderr, "Error in pthread_mutex_lock: %s\n", strerror(rc));
    if (exitrc == 0)
      exitrc = rc;
    return NULL;
  }
  NWorkersRunning += 1;
  rc = pthread_mutex_unlock(&QueueMutex);
  if (rc != 0)
  {
    fprintf(stderr, "Error in pthread_mutex_unlock: %s\n", strerror(rc));
    NWorkersRunning -= 1;
    if (exitrc == 0)
      exitrc = rc;
    return NULL;
  }

  /* Loop while there is work to do */
  while (1)
  {
    /* Wait for work */
    qP = DequeueWork();
    if (qP == NULL)
      break;
    
    /* Read this directory */
    rc = read_dir(qP->path, qP->ino);
    if (exitrc == 0)
      exitrc = rc;

    /* Free work queue element */
    free(qP);

  }

  rc = pthread_mutex_lock(&QueueMutex);
  if (rc != 0)
  {
    fprintf(stderr, "Error in pthread_mutex_lock: %s\n", strerror(rc));
    if (exitrc == 0)
      exitrc = rc;
    return NULL;
  }
  NWorkersRunning -= 1;
  rc = pthread_mutex_unlock(&QueueMutex);
  if (rc != 0)
  {
    fprintf(stderr, "Error in pthread_mutex_unlock: %s\n", strerror(rc));
    if (exitrc == 0)
      exitrc = rc;
  }

  return NULL;
}

/* Read all entries from the current directory */
int read_dir(const char *pathP, ino_t inode)
{
  int rc = 0, ind;
  const gpfs_direntx_t *direntxP;
  gpfs_fssnap_handle_t *fsP = NULL;
  gpfs_ifile_t *dirxP = NULL;
  const char *typeP;
  const char *slashP;

  /* Check if we need a "/"  */
  if (strlen(pathP) > 0)
    slashP = "/";
  else
    slashP = "";

  /* Open the file system */
  fsP = gpfs_get_fssnaphandle_by_path(RootFsDirP);
  if (fsP == NULL)
  {
    rc = errno;
    fprintf(stderr, "gpfs_get_fssnaphandle_by_path(%s): %s\n",
            RootFsDirP, strerror(rc));
    goto exit;
  }

  /* If there is only one inode in the list and is one of the reserved inode
     it cannot possibly be found, so stop the nonsense now.
     Calling tsfindinode -i 1 can be a fast test as to whether the filesystem
     is mounted and ready to use. */
  if (numInodes == 1 && Inode[0] < 3)
    goto exit;

  /* Open the directory's file */
  dirxP = gpfs_iopen(fsP, inode, O_RDONLY, NULL, NULL);
  if (dirxP == NULL)
  {
    rc = errno;
    fprintf(stderr, "gpfs_iopen(%s): %s\n", pathP, strerror(rc));
    goto exit;
  }

  /* Loop reading the directory entries */
  while (1)
  {
    rc = gpfs_ireaddir(dirxP, &direntxP);
    if (rc != 0)
    {
      int saveerrno = errno;
      fprintf(stderr, "gpfs_ireaddir(%s): rc %d %s\n",
              pathP, rc, strerror(saveerrno));
      rc = saveerrno;
      goto exit;
    }
    if (direntxP == NULL)
      break;
    
    /* Get directory entry type */
    switch (direntxP->d_type)
    {
      case GPFS_DE_DIR:
        /* Skip . and .. entries */
        if ((strcmp(direntxP->d_name,".") == 0) ||
            (strcmp(direntxP->d_name,"..") == 0))
          continue;

        typeP = "DIR";
        break;

      case GPFS_DE_REG:
        typeP = "REG";
        break;

      case GPFS_DE_LNK:
        typeP = "LNK";
        break;

      default:
        typeP = "UNK";
    }

    ind = testinode(direntxP->d_ino);
    if (ind >= 0)
    {
      int plen;
      unsigned long long dirinode = direntxP->d_ino;

      /* Print file info on stdout.
         Serialize output with all other threads */
      rc = pthread_mutex_lock(&OutputMutex);
      if (rc != 0)
      {
        fprintf(stderr, "Error in pthread_mutex_lock: %s\n", strerror(rc));
        goto exit;
      }
      /* Print: inodenumber pathname */
      plen = printf("%10llu\t%10d\t%s%s%s%s",
                    dirinode, snapIdIndex, pathP, slashP, direntxP->d_name, print0);
      InodeSeen[ind] = 1;
      rc = pthread_mutex_unlock(&OutputMutex);
      if (rc != 0)
      {
        fprintf(stderr, "Error in pthread_mutex_unlock: %s\n", strerror(rc));
        goto exit;
      }
      if (plen < 0)
      {
        rc = errno;
        goto exit;
      }
    }

    /* Check if this entry is a directory */
    if (direntxP->d_type == GPFS_DE_DIR)
    {
      /* Add this directory to the work queue */
      rc = EnqueueWork(pathP, direntxP->d_name, direntxP->d_ino);
      if (rc != 0)
      {
        if (exitrc == 0)
          exitrc = rc;
        goto exit;
      }
    }
  }
  

exit:
  /* Close open file, if necessary */
  if (dirxP)
    gpfs_iclose(dirxP);
  if (fsP)
    gpfs_free_fssnaphandle(fsP);

  return rc;
}

/* main */
int main(int argc, char *argv[])
{
  int i, rc, filearg, ind;
  struct stat statBuf;
  pthread_t pThread;
  gpfs_fssnap_handle_t *fsP;

  /* First scan for -a option. */
  for (i = 1; i < argc; i++)
  {
    if (0 == strcmp(argv[i],"-a"))
    {
      isScanAllSnapshots = 1;
      if (i + 1 < argc)
      {
        i += 1;
        fsName = argv[i];
      }
      else
        usage(argv[0]);
    }
    else if(0 == strcmp(argv[i],"-i"))
    {
      if (i + 1 < argc)
      {
        i += 1;
        inputFileName = argv[i];
      }
      else
        usage(argv[0]);
    }
    else if (0 == strcmp(argv[i],"-t"))
    {
      i += 1;
      if (i < argc)
        nThreads = atoi(argv[i]);
      else
        usage(argv[0]);
    }
  }

  /* Open the gpfs device in the beginning and keep it opened */
  rc = gpfs_lib_init(0);
  if (rc)
  {
    fprintf(stderr, "Failed to open GPFS device, please check the GPFS"
            " system status and retry the command again\n");
    return rc;
  }

  if (isScanAllSnapshots)
  {
    /* scanAllSnapshots constructs all necessary working files and
       then fork a new tsfindinode to search snapshots. */
    rc = scanAllSnapshots();
    return rc;
  }

  /* Scan for options and single directory name */
  filearg = 0;
  for (i = 1; i < argc; i++)
  {
    if (0 == strcmp(argv[i],"-i"))
    {
      i += 1;
      if (i < argc)
        getinodes(argv[i]);
      else
        usage(argv[0]);
    }
    else if (0 == strcmp(argv[i],"-t"))
    {
      i += 1;
      if (i < argc)
        nThreads = atoi(argv[i]);
      else
        usage(argv[0]);
    }
    else if (0 == strcmp(argv[i],"-s"))
    {
      i += 1;
      if (i < argc)
        snapIdIndex = atoi(argv[i]);
      else
        snapIdIndex = -1;
    }
    else if (0 == strcmp(argv[i],"-0"))
      print0 = "\0";
    else if (0 == strcmp(argv[i],"--"))
      optcheck = 0;
    else if (optcheck && *argv[i] == '-')
      usage(argv[0]);
    else if (filearg == 0)
      filearg = i;
    else
      usage(argv[0]);
  }

  /* if no directory was specified, there is nothing to do */
  if (filearg == 0 || numInodes == 0)
    usage(argv[0]);
  RootFsDirP = argv[filearg];

  /* Initialize pthread variables */
  rc = pthread_mutex_init(&QueueMutex, NULL);
  if (rc != 0)
  {
    fprintf(stderr, "Error from pthread_mutex_init: %s\n", strerror(rc));
    exit(rc);
  }
  rc = pthread_cond_init(&QueueCond, NULL);
  if (rc != 0)
  {
    fprintf(stderr, "Error from pthread_cond_init: %s\n", strerror(rc));
    exit(rc);
  }
  rc = pthread_mutex_init(&OutputMutex, NULL);
  if (rc != 0)
  {
    fprintf(stderr, "Error from pthread_mutex_init: %s\n", strerror(rc));
    exit(rc);
  }
  
  /* Get the inode number of the root directory */
  rc = stat(RootFsDirP, &statBuf);
  if (rc) 
  {
    perror(RootFsDirP);
    return rc;
  }

  /* Verify this is a directory */
  if (!S_ISDIR(statBuf.st_mode))
  {
    errno = ENOTDIR;
    perror(RootFsDirP);
    return ENOTDIR;
  }
  
  /* Get the snapId for this scan. */
  if (snapIdIndex == -1)
  {
    fsP = gpfs_get_fssnaphandle_by_path(RootFsDirP);
    if (fsP == NULL)
    {
      rc = errno;
      fprintf(stderr, "gpfs_get_fssnaphandle_by_path(%s): %s\n",
              RootFsDirP, strerror(rc));
      return rc;
    }
    snapIdIndex = gpfs_get_snapid_from_fssnaphandle(fsP);
    gpfs_free_fssnaphandle(fsP);
  }
 
  InodeSeen = (unsigned char *) malloc(maxInodes);
  if (InodeSeen == NULL)
  {
    fprintf(stderr,"Out of memory.\n");
    return ENOMEM;
  }
  memset(InodeSeen, 0, maxInodes);

  ind = testinode(statBuf.st_ino);
  if (ind >= 0)
  {
    /* Print file info on stdout */
    unsigned long long dirinode = statBuf.st_ino;
    rc = printf("%10llu\t%10d\t%s%s", dirinode, snapIdIndex, RootFsDirP, print0);
    if (rc < 0)
      return errno;
    InodeSeen[ind] = 1;
  }

  /* Add the root directory to the work queue */
  rc = EnqueueWork(RootFsDirP, "", statBuf.st_ino);
  if (rc != 0)
    return rc;

  /* Start the requested number of threads */
  for (i = 1; i < nThreads; i++)
  {
    rc = pthread_create(&pThread, NULL, ThreadBody, NULL);
    if (rc != 0)
    {
      fprintf(stderr, "Error from pthread_create(%d): %s\n", i, strerror(rc));
      exit(rc);
    }
  }

  /* This thread can work too */
  (void) ThreadBody(NULL);

  for (ind=0; ind < numInodes; ind++)
  {
    if (!InodeSeen[ind])
    {
      if (snapIdIndex == 0)
        fprintf(stderr, "%10llu\t%10d\t(notfound)\n", Inode[ind], snapIdIndex);
      else
        fprintf(stderr, 
                "%10llu\t%10d\t(notfound, fileset is not linked in the snapshot ?)\n",
                Inode[ind], snapIdIndex);
    }
  }

  /* No need to call this actually */
  gpfs_lib_term(0);

  return exitrc;
}


